@@ -39,13 +39,28 @@ if not owlready2_optimized:
 INT_DATATYPES   = { "http://www.w3.org/2001/XMLSchema#integer", "http://www.w3.org/2001/XMLSchema#byte", "http://www.w3.org/2001/XMLSchema#short", "http://www.w3.org/2001/XMLSchema#int", "http://www.w3.org/2001/XMLSchema#long", "http://www.w3.org/2001/XMLSchema#unsignedByte", "http://www.w3.org/2001/XMLSchema#unsignedShort", "http://www.w3.org/2001/XMLSchema#unsignedInt", "http://www.w3.org/2001/XMLSchema#unsignedLong", "http://www.w3.org/2001/XMLSchema#negativeInteger", "http://www.w3.org/2001/XMLSchema#nonNegativeInteger", "http://www.w3.org/2001/XMLSchema#positiveInteger" }
 FLOAT_DATATYPES = { "http://www.w3.org/2001/XMLSchema#decimal", "http://www.w3.org/2001/XMLSchema#double", "http://www.w3.org/2001/XMLSchema#float", "http://www.w3.org/2002/07/owl#real" }
 
+
+class _FakeQueue(object):
+  def __init__(self, insert_objs, insert_datas, finish):
+    self.insert_objs  = insert_objs
+    self.insert_datas = insert_datas
+    self.finish       = finish
+    
+  def put(self, args):
+    command, triples = args
+    if   command == "objs":   self.insert_objs(triples)
+    elif command == "datas":  self.insert_datas(triples)
+    elif command == "finish": return self.finish()
+    elif command == "error":  raise OwlReadyOntologyParsingError(*args)
+    
+    
 class BaseGraph(object):
   _SUPPORT_CLONING = False
   #READ_METHODS  = ["_refactor", "_new_numbered_iri", "_abbreviate", "_unabbreviate",
   #                 "get_triple_sp", "_get_data_triple_triple_sp", "get_triple_po", "_get_obj_triples_transitive_sp", "_get_obj_triples_transitive_po", "_get_obj_triples_transitive_sym", "_get_obj_triples_transitive_sp_indirect", "get_triples", "_get_data_triple_triples", "get_triples_s", "get_triples_sp", "_get_data_triple_triples_sp", "get_triples_po", "get_pred", "get_quads", "_get_triple_data_triples_sp", "_get_triple_data_triple_sp", "_get_triples_sp", "has_triple", "_has_data_triple_triple", "_del_triple", "_del_data_triple"]
   #WRITE_METHODS = ["_add_triple", "_set_triple", "_add_data_triple", "_set_data_triple"]
   
-  BASE_METHODS  = ["_refactor", "_new_numbered_iri", "_abbreviate", "_unabbreviate",
+  BASE_METHODS  = ["_refactor", "_refactor_onto", "_new_numbered_iri", "_abbreviate", "_unabbreviate",
                    
                    "_get_obj_triples_cspo_cspo", "_get_obj_triples_spo_spo", "_get_obj_triples_sp_co", "_get_obj_triples_s_po",
                    "_get_obj_triples_po_s", "_get_obj_triples_sp_o", "_get_obj_triple_sp_o", "_get_obj_triple_po_s", "_has_obj_triple_spo", "_del_obj_triple_raw_spo",
@@ -135,52 +150,56 @@ class BaseSubGraph(BaseGraph):
     format = format or _guess_format(f)
     
     if   format == "ntriples":
-      objs, datas, on_prepare_obj, on_prepare_data, insert_objs, insert_datas, new_blank, _abbreviate, on_finish = self.create_parse_func(getattr(f, "name", ""), delete_existing_triples)
-      
+      current_line = 0
       try:
-        current_line = 0
-        if owlready2_optimized:
-          owlready2_optimized.parse_ntriples(f, objs, datas, insert_objs, insert_datas, _abbreviate, new_blank, default_base)
+        queue = _FakeQueue(*self.import_triples_from_queue(None, getattr(f, "name", ""), delete_existing_triples))
+        if  owlready2_optimized:
+          owlready2_optimized.parse_ntriples(f, queue, default_base, 800000)
           
         else:
           splitter = re.compile("\s")
-          bn_src_2_sql = {}
-          
-          line = f.readline().decode("utf8")
+          objs  = []
+          datas = []
+          line  = f.readline().decode("utf8")
           while line:
             current_line += 1
             if (not line.startswith("#")) and (not line.startswith("\n")):
-              if not line.endswith("\n"): line = "%s\n" % line
-              s,p,o = splitter.split(line[:-3], 2)
+              if not line.endswith("\n"): s,p,o = splitter.split(line[:-2], 2)
+              else:                       s,p,o = splitter.split(line[:-3], 2)
               
-              if   s.startswith("<"): s = s[1:-1]
-              elif s.startswith("_"):
-                bn = bn_src_2_sql.get(s)
-                if bn is None: bn = bn_src_2_sql[s] = new_blank()
-                s = bn
+              if s.startswith("<"): s = s[1:-1]
               
               p = p[1:-1]
-            
-              if   o.startswith("<"): on_prepare_obj(s, p, o[1:-1])
-              elif o.startswith("_"):
-                bn = bn_src_2_sql.get(o)
-                if bn is None: bn = bn_src_2_sql[o] = new_blank()
-                on_prepare_obj(s, p, bn)
-              elif o.startswith('"'):
+              
+              if   o.startswith("<"):
+                objs.append((s, p, o[1:-1]))
+                if len(objs) > 800000:
+                  queue.put(("objs", objs))
+                  objs = []
+                  
+              elif o.startswith("_"): objs.append((s, p, o))
+                
+              else: #if o.startswith('"'):
                 o, d = o.rsplit('"', 1)
                 if d.startswith("^"):
                   d = d[3:-1]
                   if   d in INT_DATATYPES:   o = int  (o[1:])
                   elif d in FLOAT_DATATYPES: o = float(o[1:])
                   else:                      o = o[1:].encode("raw-unicode-escape").decode("unicode-escape")
-                else:
-                  o = o[1:].encode("raw-unicode-escape").decode("unicode-escape")
-                on_prepare_data(s, p, o, d)
-                
+                elif d.startswith("@"):      o = o[1:].encode("raw-unicode-escape").decode("unicode-escape")
+                else:                        o = o[1:].encode("raw-unicode-escape").decode("unicode-escape"); d = ""
+                datas.append((s, p, o, d))
+                if len(datas) > 800000:
+                  queue.put(("datas", datas))
+                  datas = []
+                  
             line = f.readline().decode("utf8")
+            
+          if objs:  queue.put(("objs", objs))
+          if datas: queue.put(("datas", datas))
           
-        onto_base_iri = on_finish()
-        
+        onto_base_iri = queue.put(("finish", None))
+
       except Exception as e:
         if len(self) == 0:
           self._add_obj_triple_raw_spo(self.onto.storid, rdf_type, owl_ontology)
@@ -188,35 +207,61 @@ class BaseSubGraph(BaseGraph):
           raise OwlReadyOntologyParsingError("NTriples parsing error (or unrecognized file format) in %s, line %s." % (getattr(f, "name", getattr(f, "url", "???")), current_line)) from e
         else:
           raise OwlReadyOntologyParsingError("NTriples parsing error (or unrecognized file format) in %s." % getattr(f, "name", getattr(f, "url", "???"))) from e
-          
-    elif format == "rdfxml":
-      objs, datas, on_prepare_obj, on_prepare_data, insert_objs, insert_datas, new_blank, _abbreviate, on_finish = self.create_parse_func(getattr(f, "name", ""), delete_existing_triples)
-      try:
-        if owlready2_optimized:
-          owlready2_optimized.parse_rdfxml(f, objs, datas, insert_objs, insert_datas, _abbreviate, new_blank, default_base)
-        else:
-          import owlready2.rdfxml_2_ntriples
-          owlready2.rdfxml_2_ntriples.parse(f, on_prepare_obj, on_prepare_data, new_blank, default_base)
-        onto_base_iri = on_finish()
-      except OwlReadyOntologyParsingError as e:
-        if len(self) == 0: self._add_obj_triple_raw_spo(self.onto.storid, rdf_type, owl_ontology)
-        raise e
-      
-    elif format == "owlxml":
-      objs, datas, on_prepare_obj, on_prepare_data, insert_objs, insert_datas, new_blank, _abbreviate, on_finish = self.create_parse_func(getattr(f, "name", ""), delete_existing_triples)
-      try:
-        if owlready2_optimized:
-          owlready2_optimized.parse_owlxml(f, objs, datas, insert_objs, insert_datas, _abbreviate, new_blank, default_base)
-        else:
-          import owlready2.owlxml_2_ntriples
-          owlready2.owlxml_2_ntriples.parse(f, on_prepare_obj, on_prepare_data, new_blank, default_base)
-        onto_base_iri = on_finish()
-      except OwlReadyOntologyParsingError as e:
-        if len(self) == 0: self._add_obj_triple_raw_spo(self.onto.storid, rdf_type, owl_ontology)
-        raise e
-      
+
+        
     else:
-      raise ValueError("Unsupported format %s." % format)
+      queue = None
+      def do_parse(batch_size = 30000):
+        nonlocal queue
+        try:
+          if owlready2_optimized:
+            if   format == "rdfxml": owlready2_optimized.parse_rdfxml(f, queue, default_base, batch_size)
+            elif format == "owlxml": owlready2_optimized.parse_owlxml(f, queue, default_base, batch_size)
+          else:
+            objs  = []
+            datas = []
+            def on_prepare_obj(*triple):
+              nonlocal objs
+              objs.append(triple)
+              if len(objs) > 30000: queue.put(("objs", objs));  objs = []
+            def on_prepare_data(*triple):
+              nonlocal datas
+              datas.append(triple)
+              if len(datas) > 30000: queue.put(("datas", datas)); datas = []
+            if format == "rdfxml":
+              import owlready2.rdfxml_2_ntriples
+              owlready2.rdfxml_2_ntriples.parse(f, on_prepare_obj, on_prepare_data, None, default_base)
+            else:
+              import owlready2.owlxml_2_ntriples
+              owlready2.owlxml_2_ntriples.parse(f, on_prepare_obj, on_prepare_data, None, default_base)
+              
+            if objs:  queue.put(("objs",  objs))
+            if datas: queue.put(("datas", datas))
+            
+          return queue.put(("finish", None))
+          
+        except Exception as e:
+          sys.excepthook(*sys.exc_info())
+          queue.put(("error", e.args))
+          
+      try:
+        parallel = os.path.getsize(f.name) >= 8000000
+        if parallel: import multiprocessing
+      except:
+        parallel = False
+        
+      try:
+        if parallel:
+          queue = multiprocessing.Queue()
+          multiprocessing.Process(target = do_parse).start()
+          onto_base_iri = self.import_triples_from_queue(queue, getattr(f, "name", ""), delete_existing_triples)
+        else:
+          queue = _FakeQueue(*self.import_triples_from_queue(None, getattr(f, "name", ""), delete_existing_triples))
+          onto_base_iri = do_parse(800000)
+          
+      except OwlReadyOntologyParsingError as e:
+        if len(self) == 0: self._add_obj_triple_raw_spo(self.onto.storid, rdf_type, owl_ontology)
+        raise e
     
     return onto_base_iri
   
@@ -232,13 +277,13 @@ def _guess_format(f):
     f.seek(0)
   else:
     s = f.peek(1000).lstrip()
-
+    
   if isinstance(s, str): s = s.encode("utf8")
   if s.startswith(b"\xef\xbb\xbf"): s = s[3:] # Ignore byte-order mask
   
   if not s.lstrip().startswith(b"<"): return "ntriples"
   if s[s.find(b"\n") -1] == b".": return "ntriples"
-  if s.split(b"\n", 1)[0].endswith(b"."): return "ntriples"
+  if s.strip().split(b"\n", 1)[0].endswith(b"."): return "ntriples"
   
   if (b"<!DOCTYPE Ontology" in s) or (b"<!DOCTYPE owl:Ontology" in s) or (b"<Ontology xmlns=" in s): return "owlxml"
   
@@ -322,7 +367,7 @@ def _save(f, format, graph, filter = None):
       left = x[:splitat + 1]
       
       xmln = xmlns.get(left)
-      if not xmln:
+      if xmln is None:
         splitted = left[:-1].rsplit("/", 1)
         if len(splitted) == 2:
           xmln0 = left[:-1].rsplit("/", 1)[1][:4].replace("#", "").replace(":", "")
@@ -330,14 +375,17 @@ def _save(f, format, graph, filter = None):
           xmln0 = left[:4].replace("#", "").replace(":", "")
           
         if not xmln0[0] in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ": xmln0 = "x_" + xmln0
-        xmln  = xmln0 + ":"
+        xmln  = "%s:" % xmln0
         i = 2
         while xmln in xmlns_abbbrevs: xmln = "%s%s:" % (xmln0, i) ; i += 1
-        
+
         xmlns[left] = xmln = xmln
         xmlns_abbbrevs.add(xmln)
-        
-      return xmln + x[splitat + 1:]
+
+      x = x[splitat + 1:]
+      r = "%s%s" % (xmln, x)
+      if not x[0] in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ": bad_types.add(r)
+      return r
     
     lines  = []
     liness = {}
@@ -396,7 +444,6 @@ def _save(f, format, graph, filter = None):
         l = bn_2_inner_list[current_s]
         current_s = 0
       else:
-        #l = liness.get(type) or l = lines[-1]
         l = liness.get(type)
         if l is None: l = lines[-1]
         
@@ -443,10 +490,13 @@ def _save(f, format, graph, filter = None):
       if (p == rdf_type) and (type == "rdf:Description") and (not o < 0):
         t = abbrev(o)
         if not t in bad_types:
-          type = t
-          if type.startswith("#"): type = type[1:]
-          continue
-        
+          if t.startswith("#"): t = t[1:]
+          if t[0] in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ":
+            type = t
+            continue
+          else:
+            pass
+          
       p = abbrev(p)
       if p.startswith("#"): p = p[1:]
       
@@ -511,11 +561,6 @@ def _save(f, format, graph, filter = None):
           lines[-1].extend(l)
           lines[-1].append("")
           
-    def flatten(l, deep = ""):
-      for i in l:
-        if isinstance(i, list): yield from flatten(i, "%s    " % deep)
-        else:                   yield "%s%s" % (deep, i)
-        
     decls = []
     for iri, abbrev in xmlns.items():
       if   abbrev == "":  decls.append('xml:base="%s"' % iri)
@@ -524,6 +569,28 @@ def _save(f, format, graph, filter = None):
     if base_iri.endswith("/"):
       decls.append('xmlns="%s"' % base_iri)
       
+    def flatten(l):
+      if not l: return
+      
+      c = [l]
+      p = [0]
+      while c:
+        v = c[-1][p[-1]]
+        deep = len(c)
+        
+        if p[-1] + 1 < len(c[-1]):
+          p[-1] += 1
+        else:
+          del c[-1]
+          del p[-1]
+          
+        if isinstance(v, list):
+          if v:
+            c.append(v)
+            p.append(0)
+        else:
+          yield "%s%s" % ("    " * (deep - 1), v)
+          
     f.write(b"""<?xml version="1.0"?>\n""")
     f.write(("""<rdf:RDF %s>\n\n""" % "\n         ".join(decls)).encode("utf8"))
     f.write( """\n""".join(flatten(sum(lines, []))).encode("utf8"))
