@@ -66,10 +66,10 @@ class EntityClass(type):
   def get_name(Class): return Class._name
   def set_name(Class, name):
     Class._name = name
-    Class.namespace.world._refactor(Class.storid, "%s%s" % (Class.namespace.base_iri, name))
+    Class.namespace.world._refactor(Class.storid, "%s%s" % (Class.namespace._base_iri, name))
   name = property(get_name, set_name)
   
-  def get_iri(Class): return "%s%s" % (Class.namespace.base_iri, Class._name)
+  def get_iri(Class): return "%s%s" % (Class.namespace._base_iri, Class._name)
   def set_iri(Class, new_iri):
     splitted = new_iri.rsplit("#", 1)
     if len(splitted) == 2:
@@ -108,7 +108,7 @@ class EntityClass(type):
     
   def __new__(MetaClass, name, superclasses, obj_dict):
     namespace = obj_dict.get("namespace") or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or superclasses[0].namespace
-    storid    = obj_dict.get("storid")    or namespace.world._abbreviate("%s%s" % (namespace.base_iri, name))
+    storid    = obj_dict.get("storid")    or namespace.world._abbreviate("%s%s" % (namespace._base_iri, name))
     
     if "is_a" in obj_dict:
       _is_a = [*superclasses, *obj_dict["is_a"]]
@@ -118,12 +118,12 @@ class EntityClass(type):
         _is_a = superclasses = MetaClass._find_base_classes(superclasses) or (Thing,)
       else:
         _is_a = superclasses
-        
+
     if LOADING:
       Class = namespace.world._entities.get (storid)
     else:
       Class = namespace.world._get_by_storid(storid)
-        
+      
     equivalent_to = obj_dict.pop("equivalent_to", None)
     
     if Class is None:
@@ -131,7 +131,7 @@ class EntityClass(type):
         for base in _is_a:
           if isinstance(base, Construct):
             base = base._set_ontology_copy_if_needed(namespace.ontology, _is_a)
-              
+        
       _is_a = CallbackList(_is_a, None, MetaClass._class_is_a_changed)
       obj_dict.update(
         _name          = name,
@@ -140,7 +140,7 @@ class EntityClass(type):
         is_a           = _is_a,
         _equivalent_to = None,
       )
-      
+
       Class = namespace.world._entities[storid] = _is_a._obj = type.__new__(MetaClass, name, superclasses, obj_dict)
       _cache_entity(Class)
       
@@ -162,7 +162,7 @@ class EntityClass(type):
     Class.namespace.ontology._add_obj_triple_spo(Class.storid, Class._rdfs_is_a, base.storid)
     
   def _del_is_a_triple(Class, base):
-    Class.namespace.ontology._del_obj_triple_spo(Class.storid, Class._rdfs_is_a, base.storid)
+    Class.namespace.world._del_obj_triple_spo(Class.storid, Class._rdfs_is_a, base.storid)
     
   def __init__(Class, name, bases, obj_dict):
     if "defined_class" in obj_dict:
@@ -203,7 +203,7 @@ class EntityClass(type):
     old = frozenset(old)
     
     for x in old - new:
-      Class.namespace.ontology._del_obj_triple_spo(Class.storid, Class._owl_equivalent, x.storid)
+      Class.namespace.world._del_obj_triple_spo(Class.storid, Class._owl_equivalent, x.storid)
       if isinstance(x, Construct): x._set_ontology(None)
       else: # Invalidate it
         if not x.equivalent_to._indirect is None:
@@ -230,7 +230,7 @@ class EntityClass(type):
       type.__setattr__(Class, "is_a", CallbackList(value, Class, Class.__class__._class_is_a_changed))
       Class._class_is_a_changed(old)
       return
-      
+
     type.__setattr__(Class, attr, value)
     
   def _class_is_a_changed(Class, old):
@@ -248,8 +248,9 @@ class EntityClass(type):
       Class.__bases__ = bases
     else:
       if   isinstance(Class, ThingClass):
-        Class.__bases__ = (Thing,)
-        list.insert(Class.is_a, 0, Thing)
+        if not Class is Thing:
+          Class.__bases__ = (Thing,)
+          list.insert(Class.is_a, 0, Thing)
       elif isinstance(Class, ObjectPropertyClass):
         Class.__bases__ = (ObjectProperty,)
         list.insert(Class.is_a, 0, ObjectProperty)
@@ -259,7 +260,10 @@ class EntityClass(type):
         
     for base in new - old:
       if isinstance(base, Construct):
-        base = base._set_ontology_copy_if_needed(Class.namespace.ontology, Class.is_a)
+        if Class is Thing:
+          base = base._set_ontology_copy_if_needed(CURRENT_NAMESPACES.get()[-1].ontology, Class.is_a)
+        else:
+          base = base._set_ontology_copy_if_needed(Class.namespace.ontology, Class.is_a)
       if not LOADING: Class._add_is_a_triple(base)
       
   def disjoints(Class):
@@ -293,6 +297,24 @@ class EntityClass(type):
       Class._fill_descendants(s, include_self, only_loaded, Class.namespace.world, Class.namespace.ontology)
     return s
   
+  # def descendants2(Class, include_self = True, only_loaded = False, world = None):
+  #   if (Class.namespace.world is owl_world) and (world is None):
+  #     import owlready2
+  #     world = owlready2.default_world
+  #   else:
+  #     world = Class.namespace.world
+
+  #   if only_loaded:
+  #     if include_self:
+  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)* ?? . FILTER(ISIRI(?c)) . FILTER(LOADED(?c)) }""").execute_flat((Class,)))
+  #     else:
+  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)*/rdfs:subClassOf ?? . FILTER(ISIRI(?c)) . FILTER(LOADED(?c)) }""").execute_flat((Class,)))
+  #   else:
+  #     if include_self:
+  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)* ?? . FILTER(ISIRI(?c)) }""").execute_flat((Class,)))
+  #     else:
+  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)*/rdfs:subClassOf ?? . FILTER(ISIRI(?c)) }""").execute_flat((Class,)))
+    
   def _fill_ancestors(Class, s, include_self, include_constructs):
     if include_self:
       if not Class in s:
@@ -314,14 +336,14 @@ class EntityClass(type):
         if isinstance(parent, EntityClass):
           if not parent in s:
             parent._fill_ancestors(s, True, False)
-        
+            
   def _fill_descendants(Class, s, include_self, only_loaded, world, onto):
     if include_self:
       s.add(Class)
       for equivalent in Class.equivalent_to.indirect():
         if isinstance(equivalent, Class.__class__) and not equivalent in s:
           equivalent._fill_descendants(s, True, only_loaded, world, onto)
-          
+    
     for x in world._get_obj_triples_transitive_po(Class._rdfs_is_a, Class.storid):
       if not x < 0:
         if only_loaded:
@@ -343,28 +365,15 @@ class EntityClass(type):
         world = owlready2.default_world
         
       for x, in world.graph.db.execute(
-    """SELECT q1.s FROM objs q1 WHERE q1.s > 0 and q1.p = ? AND q1.o = ?
-EXCEPT SELECT q2.s FROM objs q2 WHERE q2.p = ? and q2.o != ?""",
+          """SELECT q1.s FROM objs q1 WHERE q1.p=? AND q1.o=? AND ((q1.s > 0))
+    AND (SELECT 1 FROM objs q2 INDEXED BY index_objs_sp
+    WHERE q2.s=q1.s AND q2.p=? AND ((q2.o > 0) AND q2.o != ?)) IS NULL""",
           (rdf_type, owl_class, rdfs_subclassof, owl_thing)):
         if only_loaded:
           subclass = world._entities.get(x)
           if not subclass is None: yield subclass
         else:
           yield world._get_by_storid(x, None, ThingClass)
-          
-      #r = []
-      #for x in world._get_obj_triples_po_s(rdf_type, owl_class):
-      #  if x < 0: continue
-      #  for y in world._get_obj_triples_sp_o(x, Class._rdfs_is_a):
-      #    if (y == owl_thing) or y < 0: continue
-      #    break
-      #  else:
-      #    if only_loaded:
-      #      subclass = world._entities.get(x)
-      #      if not subclass is None: yield subclass
-      #    else:
-      #      yield world._get_by_storid(x, None, ThingClass)
-      #return r
       
     else:
       world = world or Class.namespace.world
@@ -461,7 +470,7 @@ class ThingClass(EntityClass):
     for removed in old:
       if not removed in new:
         Class.namespace.ontology._del_list(removed._bn)
-        Class.namespace.ontology._del_obj_triple_spo(Class.storid, owl_disjointunion, removed._bn)
+        Class.namespace.world._del_obj_triple_spo(Class.storid, owl_disjointunion, removed._bn)
     new2 = []
     for added in new:
       if not added in old:
@@ -474,14 +483,37 @@ class ThingClass(EntityClass):
     new._set(new2)
   disjoint_unions = property(get_disjoint_unions, set_disjoint_unions)
   
-      
+  
+  # def instances(Class, world = None):
+  #   if Class.namespace.world is owl_world:
+  #     import owlready2
+  #     world = (world or owlready2.default_world).world
+  #     if Class is Thing: return world.individuals()
+  #     return world.search(type = Class)
+  #   return Class.namespace.world.search(type = Class)
+  
   def instances(Class, world = None):
     if Class.namespace.world is owl_world:
       import owlready2
       world = (world or owlready2.default_world).world
       if Class is Thing: return world.individuals()
-      return world.search(type = Class)
-    return Class.namespace.world.search(type = Class)
+    else:
+      world = Class.namespace.world
+      
+    q = world.prepare_sparql("""SELECT DISTINCT ?i { ?i a/(rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)* ?? . }""")
+    if sqlite3.sqlite_version_info[1] < 36: # Some SQLite3 versions have a bug on recursive queries with multiple unions
+      q.sql = """
+WITH RECURSIVE prelim1_objs(s) AS (VALUES (?1) 
+UNION 
+SELECT q.s FROM objs q, prelim1_objs rec WHERE q.p IN (9,33) AND q.o=rec.s),
+
+prelim2_objs(s) AS (VALUES (?1)
+UNION
+SELECT q.o FROM objs q, prelim1_objs rec WHERE q.p=33 AND q.s=rec.s)
+
+SELECT q1.s FROM objs q1 WHERE q1.p=6 AND (q1.o IN (SELECT s FROM prelim1_objs) OR q1.o IN (SELECT s FROM prelim2_objs) ) ;
+"""
+    return list(q.execute_flat((Class,)))
   
   def direct_instances(Class, world = None):
     if Class.namespace.world is owl_world:
@@ -493,9 +525,11 @@ class ThingClass(EntityClass):
   def get_class_properties(Class):
     l = set()
     for r in _property_value_restrictions(Class, None):
-      if   r.property._class_property_some and ((r.type == VALUE) or (r.type == SOME) or ((r.type == EXACTLY) and r.cardinality >= 1) or ((r.type == MIN) and r.cardinality >= 1)):
+      if isinstance(r.property, owlready2.class_construct.Inverse): prop = r.property.property
+      else:                                                         prop = r.property
+      if   prop._class_property_some and ((r.type == VALUE) or (r.type == SOME) or ((r.type == EXACTLY) and r.cardinality >= 1) or ((r.type == MIN) and r.cardinality >= 1)):
         l.add(r.property)
-      elif r.property._class_property_only and  (r.type == ONLY):
+      elif prop._class_property_only and  (r.type == ONLY):
         l.add(r.property)
         
     for storid in Class.namespace.world._get_triples_s_p(Class.storid):
@@ -507,9 +541,11 @@ class ThingClass(EntityClass):
   def INDIRECT_get_class_properties(Class):
     l = set()
     for r in _inherited_properties_value_restrictions(Class, None, set()):
-      if   r.property._class_property_some and ((r.type == VALUE) or (r.type == SOME) or ((r.type == EXACTLY) and r.cardinality >= 1) or ((r.type == MIN) and r.cardinality >= 1)):
+      if isinstance(r.property, owlready2.class_construct.Inverse): prop = r.property.property
+      else:                                                         prop = r.property
+      if   prop._class_property_some and ((r.type == VALUE) or (r.type == SOME) or ((r.type == EXACTLY) and r.cardinality >= 1) or ((r.type == MIN) and r.cardinality >= 1)):
         l.add(r.property)
-      elif r.property._class_property_only and  (r.type == ONLY):
+      elif prop._class_property_only and  (r.type == ONLY):
         l.add(r.property)
         
     for storid in Class.namespace.world._get_triples_s_p(Class.storid):
@@ -543,7 +579,7 @@ class ThingClass(EntityClass):
     if defined_class:
       Class.namespace.ontology._set_data_triple_spod(Class.storid, owlready_defined_class, *to_literal(True))
     else:
-      Class.namespace.ontology._del_data_triple_spod(Class.storid, owlready_defined_class, "true", None)
+      Class.namespace.world._del_data_triple_spod(Class.storid, owlready_defined_class, "true", None)
   defined_class = property(get_defined_class, set_defined_class)
   
   def __getattr__(Class, attr):
@@ -624,7 +660,7 @@ class ThingClass(EntityClass):
     new     = set(new)
     removed = old - new
     inverse = Prop.inverse_property
-    
+
     if Class.defined_class:
       parents, r = Class._get_defined_construct()
       
@@ -720,13 +756,13 @@ class ThingClass(EntityClass):
             else:
               r = Prop.only(Or(only_classes))
               Class.is_a.append(r)
-            
+              
       if Prop._class_property_relation:
         if Prop._owl_type == owl_object_property:
           for v in removed:
-            Class.namespace.ontology._del_obj_triple_spo(Class.storid, Prop.storid, v.storid)
+            Class.namespace.world._del_obj_triple_spo(Class.storid, Prop.storid, v.storid)
             if inverse:
-              Class.namespace.ontology._del_obj_triple_spo(v.storid, inverse.storid, Class.storid) # Also remove inverse
+              Class.namespace.world._del_obj_triple_spo(v.storid, inverse.storid, Class.storid) # Also remove inverse
               #if isinstance(v, EntityClass): v.__dict__.pop("__%s" % inverse.python_name, None) # Remove => force reloading; XXX optimizable
               #else:                          v.__dict__.pop(inverse.python_name, None) # Remove => force reloading; XXX optimizable
               
@@ -738,7 +774,7 @@ class ThingClass(EntityClass):
             
         else: # Data prop
           for v in removed:
-            Class.namespace.ontology._del_data_triple_spod(Class.storid, Prop.storid, Class.namespace.ontology._to_rdf(v)[0], None)
+            Class.namespace.world._del_data_triple_spod(Class.storid, Prop.storid, Class.namespace.ontology._to_rdf(v)[0], None)
             
           for v in new - old:
             Class.namespace.ontology._add_data_triple_spod(Class.storid, Prop.storid, *Class.namespace.ontology._to_rdf(v))
