@@ -54,12 +54,12 @@ class Thing(metaclass = ThingClass):
   def get_name(self): return self._name
   def set_name(self, name):
     self._name = name
-    self.namespace.world._refactor(self.storid, "%s%s" % (self.namespace.base_iri, name))
+    self.namespace.world._refactor(self.storid, "%s%s" % (self.namespace._base_iri, name))
   name = property(get_name, set_name)
   
   def get_iri(self):
     if self.storid < 0: return ""
-    return "%s%s" % (self.namespace.base_iri, self._name)
+    return "%s%s" % (self.namespace._base_iri, self._name)
   def set_iri(self, new_iri):
     splitted = new_iri.rsplit("#", 1)
     if len(splitted) == 2:
@@ -71,7 +71,7 @@ class Thing(metaclass = ThingClass):
     self.namespace.world._refactor(self.storid, new_iri)
   iri = property(get_iri, set_iri)
   
-  def __new__(Class, name = None, namespace = None, **kargs):
+  def __new__(Class, name = None, namespace = None, is_a = None, **kargs):
     if name:
       if isinstance(name, Thing):
         namespace = name.namespace
@@ -81,7 +81,7 @@ class Thing(metaclass = ThingClass):
       if LOADING or (name == ""):
         already_existing = None
       else:
-        already_existing = namespace.world["%s%s" % (namespace.base_iri, name)]
+        already_existing = namespace.world["%s%s" % (namespace._base_iri, name)]
         
       if not already_existing is None:
         if not isinstance(already_existing, Class):
@@ -104,12 +104,9 @@ class Thing(metaclass = ThingClass):
           
         return already_existing
       
-    #_cache_entity(entity)
-    #return object.__new__(Class)
-    #print(namespace, kargs)
     return _cache_entity(object.__new__(Class))
   
-  def __init__(self, name = None, namespace = None, **kargs):
+  def __init__(self, name = None, namespace = None, is_a = None, **kargs):
     if   isinstance(name, int):
       is_new = True
       self.namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or self.__class__.namespace
@@ -118,31 +115,81 @@ class Thing(metaclass = ThingClass):
     elif name:
       is_new = not "storid" in self.__dict__
       if is_new: self.namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or self.__class__.namespace
-      iri = "%s%s" % (self.namespace.base_iri, name)
+      iri = "%s%s" % (self.namespace._base_iri, name)
       self._name = name
     else:
       is_new = True
       self.namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or self.__class__.namespace
-      iri = self.namespace.world._new_numbered_iri("%s%s" % (self.namespace.base_iri, self.generate_default_name()))
-      self._name = iri[len(self.namespace.base_iri):]
-    
+      iri = self.namespace.world._new_numbered_iri("%s%s" % (self.namespace._base_iri, self.generate_default_name()))
+      self._name = iri[len(self.namespace._base_iri):]
+
     if is_new:
       self.__dict__["_equivalent_to"] = None
       if isinstance(name, int): self.storid = name or self.namespace.world.graph.new_blank_node()
       else:                     self.storid = self.namespace.world._abbreviate(iri)
       self.namespace.world._entities[self.storid] = self
       if isinstance(self.__class__, FusionClass):
-        self.__dict__["is_a"] = CallbackList(self.__class__.__bases__, self, Thing._instance_is_a_changed)
+        self.__dict__["is_a"] = CallbackList(is_a or self.__class__.__bases__, self, Thing._instance_is_a_changed)
       else:
-        self.__dict__["is_a"] = CallbackList([self.__class__], self, Thing._instance_is_a_changed)
+        self.__dict__["is_a"] = CallbackList(is_a or [self.__class__], self, Thing._instance_is_a_changed)
         
       if not LOADING:
         if self.storid > 0: self.namespace.ontology._add_obj_triple_spo(self.storid, rdf_type, owl_named_individual)
         for parent in self.is_a:
           self.namespace.ontology._add_obj_triple_spo(self.storid, rdf_type, parent.storid)
           
-        for attr, value in kargs.items(): setattr(self, attr, value)
-    
+        #for attr, value in kargs.items(): setattr(self, attr, value)
+        #return
+        
+        for attr, value in kargs.items():
+          if value is None: continue
+          if attr in SPECIAL_ATTRS:
+            if   attr == "is_a":          self.is_a.reinit(value)
+            elif attr == "equivalent_to": self.set_equivalent_to(value) # Needed
+            else:                         super().__setattr__(attr, value)
+          else:
+            Prop = self.namespace.world._props.get(attr)
+            if Prop:
+              if Prop.is_functional_for(self.__class__):
+                super().__setattr__(attr, value)
+                if   Prop._owl_type == owl_object_property:
+                  try: del value.__dict__[Prop.inverse_property.python_name if Prop.inverse_property else "INVERSE_%s" % Prop.python_name]
+                  except (TypeError, KeyError): pass
+                  
+                  self.namespace.ontology._add_obj_triple_spo(self.storid, Prop.storid, value.storid)
+                  
+                elif Prop._owl_type == owl_data_property:
+                  self.namespace.ontology._add_data_triple_spod(self.storid, Prop.storid, *self.namespace.ontology._to_rdf(value))
+                  
+              else:
+                if not isinstance(value, list):
+                  if isinstance(Prop, AnnotationPropertyClass):
+                    if value is None: value = []
+                    else:             value = [value]
+                  else:
+                    raise ValueError("Property '%s' is not functional, cannot assign directly (use .append() or assign a list)." % attr)
+                  
+                if   Prop._owl_type == owl_object_property:
+                  for v in value:
+                    try: del v.__dict__[Prop.inverse_property.python_name if Prop.inverse_property else "INVERSE_%s" % Prop.python_name]
+                    except (TypeError, KeyError): pass
+                    self.namespace.ontology._add_obj_triple_spo(self.storid, Prop.storid, v.storid)
+
+                elif Prop._owl_type == owl_data_property:
+                  for v in value:
+                    self.namespace.ontology._add_data_triple_spod(self.storid, Prop.storid, *self.namespace.ontology._to_rdf(v))
+
+                else:
+                  for v in value:
+                    if hasattr(v, "storid"): self.namespace.ontology._add_obj_triple_spo(self.storid, Prop.storid, v.storid)
+                    else:                    self.namespace.ontology._add_data_triple_spod(self.storid, Prop.storid, *self.namespace.ontology._to_rdf(v))
+                      
+            elif attr.startswith("INVERSE_"): setattr(self, attr, value) # Not yet optimized
+              
+            else: super().__setattr__(attr, value)
+          
+          
+          
   def generate_default_name(self): return self.__class__.name.lower()
   
   def _get_is_instance_of(self):    return self.is_a
@@ -154,7 +201,7 @@ class Thing(metaclass = ThingClass):
     old = set(old)
     
     for base in old - new:
-      if not LOADING: self.namespace.ontology._del_obj_triple_spo(self.storid, rdf_type, base.storid)
+      if not LOADING: self.namespace.world._del_obj_triple_spo(self.storid, rdf_type, base.storid)
       if isinstance(base, Construct): base._set_ontology(None)
     bases = ThingClass._find_base_classes(self.is_a)
     if len(bases) == 1:
@@ -180,7 +227,7 @@ class Thing(metaclass = ThingClass):
     return self._equivalent_to
   
   def set_equivalent_to(self, value): self.equivalent_to.reinit(value)
-
+  
   # Cannot use property because it name-clashes with the Class similar property
   #equivalent_to = property(get_equivalent_to, set_equivalent_to)
   
@@ -189,7 +236,7 @@ class Thing(metaclass = ThingClass):
     old = frozenset(old)
     
     for x in old - new:
-      self.namespace.ontology._del_obj_triple_spo(self.storid, owl_equivalentindividual, x.storid)
+      self.namespace.world._del_obj_triple_spo(self.storid, owl_equivalentindividual, x.storid)
       if isinstance(x, Construct): x._set_ontology(None)
       else: # Invalidate it
         if x.equivalent_to._indirect:
@@ -260,24 +307,22 @@ class Thing(metaclass = ThingClass):
       if Prop:
         if Prop.is_functional_for(self.__class__):
           if   Prop._owl_type == owl_object_property:
+            if Prop.inverse_property: inverse_python_name = Prop.inverse_property.python_name
+            else:                     inverse_python_name = "INVERSE_%s" % Prop.python_name
             old_value = self.__dict__.get(attr, None)
-            #if Prop.inverse_property and (not old_value is None):
-            #  old_value.__dict__.pop(Prop.inverse_property.python_name, None) # Remove => force reloading; XXX optimizable
-            #  self.namespace.ontology._del_obj_triple_spo(old_value.storid, Prop.inverse_property.storid, self.storid) # Also remove inverse
             if not old_value is None:
+              old_value.__dict__.pop(inverse_python_name, None) # Remove => force reloading; XXX optimizable
               if Prop.inverse_property:
-                old_value.__dict__.pop(Prop.inverse_property.python_name, None) # Remove => force reloading; XXX optimizable
-                self.namespace.ontology._del_obj_triple_spo(old_value.storid, Prop.inverse_property.storid, self.storid) # Also remove inverse
-              else:
-                old_value.__dict__.pop("INVERSE_%s" % Prop.python_name, None) # Remove => force reloading; XXX optimizable
+                self.namespace.world._del_obj_triple_spo(old_value.storid, Prop.inverse_property.storid, self.storid) # Also remove inverse
                 
             super().__setattr__(attr, value)
             
             if value is None:
-              self.namespace.ontology._del_obj_triple_spo(self.storid, Prop.storid, None)
+              self.namespace.world._del_obj_triple_spo(self.storid, Prop.storid, None)
             else:
               self.namespace.ontology._set_obj_triple_spo(self.storid, Prop.storid, value.storid)
-              if Prop.inverse_property: value.__dict__.pop(Prop.inverse_property.python_name, None) # Remove => force reloading; XXX optimizable
+              try: del value.__dict__[inverse_python_name]
+              except (TypeError, KeyError): pass
               
           elif Prop._owl_type == owl_data_property:
             old_value = self.__dict__.get(attr, None)
@@ -285,7 +330,7 @@ class Thing(metaclass = ThingClass):
             super().__setattr__(attr, value)
             
             if value is None:
-              self.namespace.ontology._del_data_triple_spod(self.storid, Prop.storid, None, None)
+              self.namespace.world._del_data_triple_spod(self.storid, Prop.storid, None, None)
             else:
               self.namespace.ontology._set_data_triple_spod(self.storid, Prop.storid, *self.namespace.ontology._to_rdf(value))
               
@@ -319,7 +364,7 @@ class Thing(metaclass = ThingClass):
           raise ValueError("Property '%s' do not exist, cannot use its INVERSE." % attr[8:])
         
       else: super().__setattr__(attr, value)
-        
+      
   def _get_instance_possible_relations(self, ignore_domainless_properties = False):
     for Prop in self.namespace.world._reasoning_props.values():
       all_domains = set(Prop.domains_indirect())
@@ -365,43 +410,53 @@ class Thing(metaclass = ThingClass):
 
 class Nothing(Thing): pass
 
+class NamedIndividual(Thing): pass
+
+
 class FusionClass(ThingClass):
   ontology = anonymous
   
-  _CACHES         = {}
-  _FUSION_CLASSES = {}
-
   def __repr__(self): return "<FusionClass %s>" % ", ".join(repr(c) for c in self.__bases__)
   
   @staticmethod
   def _get_fusion_class(Classes0):
-    key = frozenset(Classes0)
-    Class = FusionClass._CACHES.get(key)
+    Classes0 = tuple(sorted(Classes0, key = lambda Class: Class.__name__))
+    Class = Classes0[0].namespace.world._fusion_class_cache.get(Classes0)
     if Class: return Class
     
-    Classes = _keep_most_specific(Classes0, consider_equivalence = False)
+    Classes = _keep_most_specific(Classes0, consider_equivalence = True)
     try:
-      fusion_class = FusionClass._create_fusion_class(key, Classes)
+      fusion_class = FusionClass._create_fusion_class(Classes0, Classes)
       
     except: # Too complex hierarchy => do not consider equivalent classes
-      Classes = _keep_most_specific(Classes0, consider_equivalence = True)
-      fusion_class = FusionClass._create_fusion_class(key, Classes)
+      Classes = _keep_most_specific(Classes0, consider_equivalence = False)
+      fusion_class = FusionClass._create_fusion_class(Classes0, Classes)
       
     return fusion_class
   
   @staticmethod
-  def _create_fusion_class(key, Classes):
+  def _create_fusion_class(Classes0, Classes):
+    global _NEXT_FUSION_CLASS_ID
+    
     if len(Classes) == 1:
-      FusionClass._CACHES[key] = Class = tuple(Classes)[0]
+      Class = tuple(Classes)[0]
+      Class.namespace.world._fusion_class_cache[Classes0] = Class
       return Class
     
     Classes = tuple(sorted(Classes, key = lambda Class: Class.__name__))
-    if Classes in FusionClass._FUSION_CLASSES: return FusionClass._FUSION_CLASSES[Classes]
-    name = "_AND_".join(Class.__name__ for Class in Classes)
+    fusion_class = Classes0[0].namespace.world._fusion_class_cache.get(Classes)
+    if fusion_class: return fusion_class
+    #name = "_AND_".join(Class.__name__ for Class in Classes) # Causes name clashes if several world have Fusion Class with the same OWL classes and thus the same names
+    name = "FusionClass%s" % _NEXT_FUSION_CLASS_ID
+    _NEXT_FUSION_CLASS_ID += 1
     
     with anonymous: # Force triple insertion into anonymous
-      fusion_class = FusionClass(name, Classes, { "namespace" : anonymous })
-      
-    FusionClass._FUSION_CLASSES[Classes] = FusionClass._CACHES[key] = fusion_class
+      try:
+        fusion_class = FusionClass(name, Classes, { "namespace" : anonymous })
+      except TypeError:
+        fusion_class = ThingClass(name, Classes, { "namespace" : anonymous })
+        
+    Classes0[0].namespace.world._fusion_class_cache[Classes0] = Classes0[0].namespace.world._fusion_class_cache[Classes] = fusion_class
     return fusion_class
   
+_NEXT_FUSION_CLASS_ID = 1
